import time
import functools
import threading
import collections


class RateLimiter:
    def __init__(self, max_calls, period):
        self.calls = collections.deque()

        self.period = period
        self.max_calls = max_calls
        self._lock = threading.Lock()

    def __call__(self, f):
        @functools.wraps(f)
        def wrapped(*args, **kwargs):
            with self:
                return f(*args, **kwargs)
        return wrapped

    def __enter__(self):
        with self._lock:
            if len(self.calls) >= self.max_calls:
                until = time.time() + self.period - self._timespan
                sleeptime = until - time.time()
                if sleeptime > 0:
                    time.sleep(sleeptime)
            return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        with self._lock:
            self.calls.append(time.time())

            while self._timespan >= self.period:
                self.calls.popleft()

    @property
    def _timespan(self):
        return self.calls[-1] - self.calls[0]


import datetime as dt
from typing import Optional, Union

import requests

from .models import (ChartBarsResponseModel,
                     TradeHistoryResponseModel, TokenPair)
from .ratelimit import RateLimiter

HISTORY_URL = "https://io6.dexscreener.io/u/trading-history/recent/{network}/{address}"
CHART_BAR_URL = "https://io5.dexscreener.io/u/chart/bars/{network}/{address}?{daterange}&res=15&cb=2"


class BaseClient:
    base_url = "https://api.dexscreener.io/latest"

    def __init__(self):
        self._limiter = RateLimiter(100, 60)  # 100 requests per 60 seconds

    def request(self, method, url, **kwargs) -> Union[list, dict]:
        with self._limiter:
            r = requests.request(method, url, **kwargs)

            return r.json()

    @staticmethod
    def _create_daterange_string(from_: dt.datetime, to: dt.datetime):
        return f"from={int(from_.timestamp() * 1_000)}&to={int(to.timestamp() * 1_000)}"


class DexscreenerClient(BaseClient):
    def get_token_pair(self, chain: str, address: str) -> TokenPair:
        """
        Fetch a pair on the provided chain id

        https://api.dexscreener.io/latest/dex/pairs/bsc/0x7213a321F1855CF1779f42c0CD85d3D95291D34C

        :param chain: Chain id
        :param address: Token address
        :return:
            Response as TokenPair model
        """
        url = f"{self.base_url}/dex/pairs/{chain}/{address}"

        resp = self.request("GET", url)

        return TokenPair.parse_obj(resp["pair"])

    def get_token_pairs(self, address: str) -> list[TokenPair]:
        """
        Get pairs matching base token address

        https://api.dexscreener.io/latest/dex/tokens/0x2170Ed0880ac9A755fd29B2688956BD959F933F8

        :param address: Token address
        :return:
            Response as list of TokenPair model
        """
        url = f"{self.base_url}/dex/tokens/{address}"

        resp = self.request("GET", url)

        return [TokenPair.parse_obj(pair) for pair in resp["pairs"]]

    def search_pairs(self, search_query: str) -> list[TokenPair]:
        """
        Search for pairs matching query

        https://api.dexscreener.io/latest/dex/tokens/0x2170Ed0880ac9A755fd29B2688956BD959F933F8

        :param search_query: query (e.g.: WBTC or WBTC/USDC)
        :return:
            Response as list of TokenPair model
        """
        url = f"{self.base_url}/dex/search/?q={search_query}"

        resp = self.request("GET", url)

        return [TokenPair.parse_obj(pair) for pair in resp["pairs"]]

    def chart_bars(
        self,
        network: str,
        address: str,
        from_: dt.datetime,
        to: dt.datetime
    ) -> Optional[ChartBarsResponseModel]:
        """
        Fetch data from the 'chart/bars/' endpoint

        :param network: Network name as it appears in the URL
        :param address: Pair address
        :param from_: Start date for the chart bars
        :param to: End date for the chart bars

        :return:
            Return the response model
        """
        url = CHART_BAR_URL.format(
            network=network,
            address=address,
            daterange=self._create_daterange_string(from_, to)
        )

        resp = self.request("GET", url)

        return ChartBarsResponseModel.parse_obj(resp)

    def recent_trade_history(self, network: str, address: str) -> Optional[TradeHistoryResponseModel]:
        """
        Fetch data from the 'trading-history/recent/' endpoint

        :param network: Network name as it appears in the URL
        :param address: Pair address

        :return:
            Return the response model
        """
        url = HISTORY_URL.format(
            network=network,
            address=address
        )

        resp = self.request("GET", url)

        return TradeHistoryResponseModel.parse_obj(resp)



from pydantic import BaseModel, Field
from typing import Optional
import datetime as dt


class TimestampDatetime:

    @classmethod
    def __get_validators__(cls):
        yield lambda val: dt.datetime.utcfromtimestamp(val / 1000.0) if val is not None else val


class BaseToken(BaseModel):
    address: str
    name: str
    symbol: str


class QuoteToken(BaseModel):
    symbol: str


class TransactionCount(BaseModel):
    buys: int
    sells: int


class PairTransactionCounts(BaseModel):
    m5: TransactionCount
    h1: TransactionCount
    h6: TransactionCount
    h24: TransactionCount


class _TimePeriodsFloat(BaseModel):
    m5: Optional[float] = 0.0
    h1: Optional[float] = 0.0
    h6: Optional[float] = 0.0
    h24: Optional[float] = 0.0


class VolumePeriods(_TimePeriodsFloat):
    ...


class PriceChangePeriods(_TimePeriodsFloat):
    ...


class Liquidity(BaseModel):
    usd: Optional[float] = None
    base: float
    quote: float


class TokenPair(BaseModel):
    chain_id: str = Field(..., alias="chainId")
    dex_id: str = Field(..., alias="dexId")
    url: str
    pair_address: str = Field(..., alias="pairAddress")
    base_token: BaseToken = Field(..., alias="baseToken")
    quote_token: QuoteToken = Field(..., alias="quoteToken")
    price_native: float = Field(..., alias="priceNative")
    price_usd: Optional[float] = Field(None, alias="priceUsd")
    transactions: PairTransactionCounts = Field(..., alias="txns")
    volume: VolumePeriods
    price_change: PriceChangePeriods = Field(..., alias="priceChange")
    liquidity: Optional[Liquidity] = None
    fdv: Optional[float] = 0.0
    pair_created_at: Optional[TimestampDatetime] = Field(None, alias="pairCreatedAt")


# = Unofficial endpoint models

class NumberField(int):

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        v = v.replace(",", "")

        return float(v)


# Trade History Models #

class TradeHistoryEntryModel(BaseModel):
    block: int = Field(..., alias="blockNumber")
    block_timestamp: int = Field(..., alias="blockTimestamp")
    txn_hash: str = Field(..., alias="txnHash")
    log_index: int = Field(..., alias="logIndex")
    type: str = Field(..., alias="type")
    price_usd: NumberField = Field(..., alias="priceUsd")
    volume_usd: NumberField = Field(..., alias="volumeUsd")
    token0_amount: NumberField = Field(..., alias="amount0")
    token1_amount: NumberField = Field(..., alias="amount1")


class TradeHistoryResponseModel(BaseModel):
    schema_verion: str = Field(..., alias="schemaVersion")

    base_token_symbol: str = Field([], alias="baseTokenSymbol")
    quote_token_symbol: str = Field([], alias="quoteTokenSymbol")
    trade_history: list[TradeHistoryEntryModel] = Field([], alias="tradingHistory")


# Chart Bars #

class ChartBarModel(BaseModel):
    timestamp: int = Field(..., alias="timestamp")
    open: NumberField = Field(..., alias="open")
    open_usd: NumberField = Field(..., alias="openUsd")
    high: NumberField = Field(..., alias="high")
    high_usd: NumberField = Field(..., alias="highUsd")
    low: NumberField = Field(..., alias="low")
    low_usd: NumberField = Field(..., alias="lowUsd")
    close: NumberField = Field(..., alias="close")
    close_usd: NumberField = Field(..., alias="closeUsd")
    volume_usd: NumberField = Field(..., alias="volumeUsd")


class ChartBarsResponseModel(BaseModel):
    schema_verion: str = Field(..., alias="schemaVersion")
    bars: list[ChartBarModel] = Field([], alias="bars")